/*
 *  This file is part of Siena, a wide-area event notification system.
 *  See http://www.cs.colorado.edu/serl/siena/
 *  
 *  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
 *  See the file AUTHORS for full details. 
 *  
 *  Copyright (C) 1998-2002 University of Colorado
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 *  USA, or send email to serl@cs.colorado.edu.
 *  
 *  
 *  $Id: fwd.y,v 1.17 2002/09/29 15:02:47 carzanig Exp $
 */
%{
#include "fwdconf.h"

#include <cstring>
#include <iostream>
#include <string>
#include <set>
#include <map>
#include <assert.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/resource.h>

#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#else
# error you need unistd.h
#endif

#include "yyfwdtypes.h"
#include "fwd_table.h"

#define YYERROR_VERBOSE
#define YYDEBUG 1

int yyerror(char *);

/** resource meter
 *
 *  measures CPU usage for both user and system time
 **/
class Timer {
public:
    void start();
    void stop();

    long elapsed_sys_msec();
    long elapsed_usr_msec();

private:
    struct rusage s_start;
    struct rusage s_stop;
};

inline void Timer::start() {
    getrusage(RUSAGE_SELF, &s_start);
}

inline void Timer::stop() {
    getrusage(RUSAGE_SELF, &s_stop);
}

inline long Timer::elapsed_usr_msec() {
    long secs = s_stop.ru_utime.tv_sec - s_start.ru_utime.tv_sec;
    long msec = (s_stop.ru_utime.tv_usec - s_start.ru_utime.tv_usec) / 1000;
    return secs * 1000 + msec;
}

inline long Timer::elapsed_sys_msec() {
    long secs = s_stop.ru_stime.tv_sec - s_start.ru_stime.tv_sec;
    long msec = (s_stop.ru_stime.tv_usec - s_start.ru_stime.tv_usec) / 1000;
    return secs * 1000 + msec;
}

class ICounter : public IProcessor {
public:
    ICounter() : c(0) { }
    virtual bool output(if_t);

    unsigned long get_count() { return c; }
private:
    unsigned long  c;
};

bool ICounter::output(if_t x) {
    ++c;
    return false;
}

class IPrinter : public ICounter {
public:
    IPrinter(): ICounter() {};
    virtual bool output(if_t);
    void flush(ostream & os);

 private:
    set<if_t> iset;
};

bool IPrinter::output(if_t x) {
    ICounter::output(x);
    iset.insert(x);
    return false;
}

void IPrinter::flush(ostream & os) {
    os << "->";
    set<if_t>::const_iterator i;
    for(i = iset.begin(); i != iset.end(); ++i)
	cout << ' ' << *i;
    os << endl;
}

 map<int, sx_predicate *> testing_fwd_table;

 bool testing = false; 
 bool stats_only = false; 
 bool silent = false; 

 unsigned long i_counter = 0;
 unsigned long f_counter = 0;
 unsigned long c_counter = 0;
 unsigned long m_counter = 0;
 unsigned long a_counter = 0;
 unsigned long x_counter = 0;
 unsigned long xx_counter = 0;

 Timer T;

 FwdTable FT;
%}

%pure_parser

/*
 * Keywords & other lex tokens
 */
%token <str> ID STR
%token <int_v> INT
%token <bool_v> BOOL
%token <double_v> DOUBLE
%token AND OR LT GT EQ NE PF SF SS
%token IFCONFIG SELECT 
%token PRINTSTATS STATSONLY TIMERSTART TIMERSTOP 
%token OUTPUTON OUTPUTOFF PREPROCESSLIMIT TESTING CLEAR
/*
 * some semantic types for non-terminals
 */
%type <value>		Value
%type <constraint>	Constraint
%type <message>		Message
%type <predicate>	Predicate
%type <op>		Op

%start StmtList

%%
/*
 * Grammar rules
 */
StmtList: Stmt | StmtList Stmt
;

Stmt: IfConfig | Select 
 | PRINTSTATS	{ fwd_parser_print_statistics(); }
 | CLEAR	{ FT.clear(); }
 | STATSONLY	{ stats_only = true; }
 | TIMERSTART	{ T.start(); }
 | TIMERSTOP	{ T.stop(); }
 | OUTPUTON	{ silent = false; }
 | OUTPUTOFF	{ silent = true; }
 | PREPROCESSLIMIT EQ INT { FT.set_preprocess_rounds($3); }
 | TESTING	{ testing = true; }
;

IfConfig: IFCONFIG INT Predicate 
   {
       if (! stats_only) {
	   if ( testing ) {
	       pair<map<int, sx_predicate *>::iterator, bool> i;
	       i = testing_fwd_table.insert(map<int, sx_predicate *>::value_type($2, $3));
	       if (!i.second) {
		   delete((*i.first).second);
		   (*i.first).second = $3;
	       }
	   } else {
	       FT.ifconfig($2, *$3);
	       delete($3);
	   }
       }
       ++i_counter;
   }
;

Predicate: Constraint 
   {
       if (! stats_only) {
	   $$ = new sx_predicate();
	   $$->clear();
	   $$->push_back();
	   $$->back().add_constraint($1);
       }
       ++f_counter;
   }
 | Predicate AND Constraint
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->back().add_constraint($3);
       }
   }
 | Predicate OR Constraint
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->push_back();
	   $$->back().add_constraint($3);
	   ++f_counter;
       }
   }
;

Constraint: ID Op Value
   {
       if (! stats_only) {
	   $$ = sx_constraint($1, $2, $3);
       }
       ++c_counter;
   }
;

Op: LT { $$ = SX_lt; }
 |  GT { $$ = SX_gt; }
 |  EQ { $$ = SX_eq; }
 |  NE { $$ = SX_ne; }
 |  PF { $$ = SX_pf; }
 |  SF { $$ = SX_sf; }
 |  SS { $$ = SX_ss; }
;

Value: INT { $$ = $1; }
 |  DOUBLE { $$ = $1; }
 |  BOOL { $$ =  $1; }
 |  STR  { $$ =  $1; }
;

Select: SELECT Message 
   {
       if (! stats_only) {
	   if (silent) {
	       ICounter cc;
	       FT.match(*$2, cc);
	       if (cc.get_count() > 0) ++x_counter;
	       xx_counter += cc.get_count();
	   } else {
	       if (testing) {
		   map<int, sx_predicate *>::const_iterator i;
		   cout << "->";
		   for(i = testing_fwd_table.begin(); i != testing_fwd_table.end(); ++i)
		       if (covers(*(*i).second, *$2)) 
			   cout << ' ' << (*i).first;
		   cout << endl;
	       } else {
		   IPrinter ccc;
		   FT.match(*$2, ccc);
		   if (ccc.get_count() > 0) ++x_counter;
		   xx_counter += ccc.get_count();
		   ccc.flush(cout);
	       }
	   }
	   delete($2);
	   ++m_counter;
       }
   }
;

Message: ID EQ Value
   { 
       if (! stats_only) {
	   $$ = new sx_message();
	   $$->clear();
	   (*$$)[$1] = $3;
       }
       ++a_counter;
   }
  | Message ID EQ Value
   {
       if (! stats_only) {
	   $$ = $1;
	   (*$$)[$2] = $4;
       }
       ++a_counter;
   }
;

%%

int yyerror(char *s) 
{
    cerr << yyfwdfname << ':' << yyfwdnlin << ": " << s << endl;
    return -1;
};

int fwd_parser_run()
{
    fwd_parser_clear_statistics();
    return yyparse();
}

int fwd_parser_stats_only()
{
    int res;
    stats_only = true;
    res = yyparse();
    fwd_parser_print_statistics();
    return res;
}

void fwd_parser_timer_start() 
{
    T.start();
}

void fwd_parser_timer_stop() 
{
    T.stop();
}

void fwd_parser_print_statistics() 
{
    cout << "I=" << i_counter 
	 << " F=" << f_counter 
	 << " C=" << c_counter 
	 << " M=" << m_counter 
	 << " A=" << a_counter 
	 << " MX=" << x_counter 
	 << " MXX=" << xx_counter 
	 << " T=" << T.elapsed_usr_msec()
	 << " S=" << FT.bytesize()
	 << endl;
}

void fwd_parser_clear_statistics() 
{
    i_counter = 0;
    f_counter = 0;
    c_counter = 0;
    m_counter = 0;
    a_counter = 0;
    x_counter = 0;
    xx_counter = 0;
}
